import time
import shutil
import os

from numpy import inf


def action_confirm(msg):
    import readchar
    res = ''
    options = ['y', 'n', 'Y', 'N']
    print(msg + " Press y/n")
    while res not in options:
        res = readchar.readchar()
    return res == 'y' or res == 'Y'

def auto_rename(path):
    count = 1
    new_path = path
    while True:
        if not os.path.exists(new_path):
            return new_path
        file_name = os.path.basename(path)
        name, ext = file_name.split('.')
        new_file_name = '{}_{}.{}'.format(name, count, ext)
        new_path = os.path.join(os.path.dirname(path), new_file_name)
        count += 1


def format_num(n, fmt='{0:.3g}'):
    f = fmt.format(n).replace('+0', '+').replace('-0', '-')
    n = str(n)
    return f if len(f) < len(n) else n


def text_divider(sym, text='', ncols=None):
    if ncols is None:
        ncols = shutil.get_terminal_size()[0]
    left = ncols // 2
    right = ncols - left
    divider = sym*(left-len(text)) + text + sym*right
    return divider


class CheckpointCleaner:
    def __init__(self, root, keep='all'):
        """ This class provide the ability to control the storage size of checkpoints. 

            Args:
                root: root dir path of checkpoints
                keep: the mode to control storage. Defaults to 'all'. The valid choice includes
                      `all`: keep all checkpoints, `latest_#`: keep latest n checkpoints
        """
        self.root = root
        self.mode = keep
        self._check_mode()
    
    def _check_mode(self):
        if self.mode.startswith('latest'):
            n = int(self.mode.split('_')[-1])
        elif self.mode == 'all':
            return 
        else:
            raise ValueError(f'Invalid mode {self.mode}')
            
            
    def clean(self):
        file_names = os.listdir(self.root)
        file_names = filter(lambda x: x.endswith('.pth'), file_names)
        file_names = filter(lambda x: x.startswith('model-epoch'), file_names)
        file_names = [os.path.join(self.root, n) for n in file_names]
        file_names = sorted(file_names, key=lambda x: os.path.getctime(x), reverse=True)

        if self.mode == 'all':
            return
        elif self.mode.startswith('latest'):
            n = int(self.mode.split('_')[-1])
            for path in file_names[n:]:
                os.remove(path)


class Timer:
    def __init__(self):
        self._start_time = 0

    def tic(self):
        self._start_time = time.time()

    def tok(self):
        now = time.time()
        used = int(now - self._start_time)
        second = used % 60
        used = used // 60
        minutes = used % 60
        used = used // 60
        hours = used
        self._start_time = time.time()
        return "{}:{}:{}".format(hours, minutes, second)


class MetricTracker:
    def __init__(self):
        self._data = {}
        self.reset()

    def reset(self):
        self._data = {}

    def update(self, key, value, n=1):
        if key not in self._data.keys():
            self._data[key] = {'total': 0, 'count': 0}
        self._data[key]['total'] += value * n
        self._data[key]['count'] += n

    def avg(self, key):
        return self._data[key]['total'] / self._data[key]['count']

    def result(self):
        return {k: self._data[k]['total'] / self._data[k]['count'] for k in self._data.keys()}

    def summary(self):
        items = ['{}: {:.8f}'.format(k, v) for k, v in self.result().items()]
        return ' '.join(items)


class PerformanceMonitor:
    def __init__(self, mnt_mode, early_stop_threshold=0.1):
        self.mnt_mode = mnt_mode
        self.early_stop_threshold = early_stop_threshold

        assert self.early_stop_threshold > 0, 'early_stop_threshold should be greater than 0'
        assert self.mnt_mode in ['min', 'max']

        self.reset()

    def update(self, metric, info=None):
        improved = (self.mnt_mode == 'min' and self.mnt_best - metric >= self.early_stop_threshold) or \
                   (self.mnt_mode == 'max' and metric - self.mnt_best >= self.early_stop_threshold)
        self.best = False
        if improved:
            self.mnt_best = metric
            self.not_improved_count = 0
            self.best = True
            self.mnt_best_info = info
        else:
            self.not_improved_count += 1

    def is_best(self):
        return self.best == True

    def should_early_stop(self, not_improved_count=5):
        return self.not_improved_count >= not_improved_count

    def reset(self):
        self.not_improved_count = 0
        self.mnt_best = inf if self.mnt_mode == 'min' else -inf
        self.best = False
        self.mnt_best_info = None

    def state_dict(self):
        return {'not_improved_count': self.not_improved_count, 'mnt_best': self.mnt_best, 
                'best': self.best, 'mnt_best_info': self.mnt_best_info}

    def load_state_dict(self, states):
        self.not_improved_count = states['not_improved_count']
        self.mnt_best = states['mnt_best']
        self.best = states['best']
        self.mnt_best_info = states.get('mnt_best_info')


import threading

class LockedIterator(object):
    def __init__(self, it):
        self.lock = threading.Lock()
        self.it = it.__iter__()

    def __iter__(self): return self

    def __next__(self):
        self.lock.acquire()
        try:
            return next(self.it)
        finally:
            self.lock.release()


import torch.utils.data as data
from PIL import Image
import os

IMG_EXTENSIONS = [
    '.jpg', '.JPG', '.jpeg', '.JPEG',
    '.png', '.PNG', '.ppm', '.PPM', '.bmp', '.BMP', '.tiff'
]


def is_image_file(filename):
    return any(filename.endswith(extension) for extension in IMG_EXTENSIONS)


def make_dataset(dir):
    images = []
    assert os.path.isdir(dir), '%s is not a valid directory' % dir

    for root, _, fnames in sorted(os.walk(dir, followlinks=True)):
        for fname in fnames:
            if is_image_file(fname):
                path = os.path.join(root, fname)
                images.append(path)

    return images


def filter(imgs, loader, min_size=None):
    if min_size is None:
        return imgs
    filtered_imgs = []
    for img in imgs:
        w, h = loader(img).size
        if w < min_size or h < min_size:
            continue
        filtered_imgs.append(img)
    return filtered_imgs


def default_loader(path):
    return Image.open(path).convert('RGB')


def gray_loader(path):
    return Image.open(path).convert('L')


class ImageFolder(data.Dataset):
    def __init__(self, root, transform=None, return_paths=False,
                 loader=default_loader, min_size=None):
        imgs = make_dataset(root)
        if len(imgs) == 0:
            raise(RuntimeError("Found 0 images in: " + root + "\n"
                               "Supported image extensions are: " +
                               ",".join(IMG_EXTENSIONS)))
        imgs = filter(imgs, loader=loader, min_size=min_size)
        self.root = root
        self.imgs = imgs
        self.transform = transform
        self.return_paths = return_paths
        self.loader = loader

    def __getitem__(self, index):
        path = self.imgs[index]
        img = self.loader(path)
        if self.transform is not None:
            img = self.transform(img)
        if self.return_paths:
            return img, path
        else:
            return img

    def __len__(self):
        return len(self.imgs)


class ImageFolderDataset(ImageFolder):
    def __init__(
            self, root, common_transform=None, input_transform=None, target_transform=None, loader=default_loader,
            min_size=None):
        super().__init__(root, loader=loader, min_size=min_size)
        self.common_transform = common_transform
        self.input_transform = input_transform
        self.target_transform = target_transform

    def __getitem__(self, index):
        img = super().__getitem__(index)
        input = output = self.common_transform(img)
        if self.input_transform is not None:
            input = self.input_transform(input)
        if self.target_transform is not None:
            output = self.target_transform(output)
        return input, output



from typing import NamedTuple
from pathlib import Path
import os


from tqdm import tqdm
from qqdm import qqdm
from colorama import Fore, init

from ..logging.logger import Logger
from .module import Module
from ._util import Timer, MetricTracker, PerformanceMonitor, CheckpointCleaner, text_divider, format_num, action_confirm

init(autoreset=True)

timer = Timer()


class Experiment:
    def __init__(self, save_dir):
        self.save_dir = Path(save_dir)
        self.log_dir = self.save_dir / 'log'
        self.ckpt_dir = self.save_dir / 'ckpt'

    def create(self):
        self.log_dir.mkdir(exist_ok=True, parents=True)
        self.ckpt_dir.mkdir(exist_ok=True, parents=True)
        return self


class EngineConfig(NamedTuple):
    max_epochs: int = 10
    log_step: int = 20
    log_img_step: int = 20
    valid_log_img_step: int = 1

    save_per_epoch: int = 1
    valid_per_epoch: int = 1

    mnt_mode: str = 'min'
    mnt_metric: str = 'loss'
    enable_early_stop: bool = False
    early_stop_threshold: float = 0.01
    early_stop_count: int = 5

    pbar: str = 'tqdm'
    num_fmt: str = '{:8.5g}'
    ckpt_save_mode: str = 'all'
    enable_tensorboard: bool = False


class Engine:
    def __init__(self, module: Module, save_dir, **kwargs):
        self.module = module
        self.experiment = Experiment(save_dir).create()
        self.cfg = EngineConfig(**kwargs)
        self.logger = Logger(self.experiment.log_dir, self.cfg.enable_tensorboard)
        self.monitor = PerformanceMonitor(self.cfg.mnt_mode, self.cfg.early_stop_threshold)
        self.ckpt_cleaner = CheckpointCleaner(self.experiment.ckpt_dir, keep=self.cfg.ckpt_save_mode)
        self.start_epoch = 1
        self.debug_mode = False

        self.ckpt_cleaner.clean()

    def set_debug_mode(self):
        self.debug_mode = True

    def resume(self, model_name, base_dir=None):
        ckpt_dir = self.experiment.ckpt_dir if base_dir is None else Experiment(base_dir).ckpt_dir
        resume_path = ckpt_dir / 'model-{}.pth'.format(model_name)
        assert resume_path.exists(), '{} not found'.format(resume_path)
        self._resume_checkpoint(resume_path)
        return self

    def train(self, train_loader, valid_loader=None):
        """
        Full training logic
        """
        for epoch in range(self.start_epoch, self.cfg.max_epochs + 1):
            self._show_divider('=', text='Epoch[{}]'.format(epoch))
            try:
                self._train(train_loader, epoch, valid_loader)
            except KeyboardInterrupt:
                msg = Fore.RED + "Oops!" + Fore.RESET + " Training was terminated by Ctrl-C. "
                "Do you want to save the latest checkpoint?"
                if action_confirm(msg):
                    save_path = self._save_checkpoint(epoch, f'interrupt_{epoch}')
                    self.logger.warning("Saving interrupted checkpoint: {} ...".format(save_path))
                return

            if self.cfg.enable_early_stop and self.monitor.should_early_stop(self.cfg.early_stop_count):
                self.logger.warning("Validation performance didn\'t improve for {} epochs. "
                                    "Training stops.".format(self.monitor.not_improved_count))
                break

    # TODO: use this function to support mannuly control training dataloader outside of engine
    def _train(self, train_loader, epoch, valid_loader=None):
        timer.tic()
        result = self._train_epoch(epoch, train_loader)
        self.epoch = epoch

        # ------------------ Save logged informations into log dict ------------------ #
        log = {'epoch': epoch, 'time': timer.tok()}
        log.update(result)

        # ------------------ Print logged informations to the screen ----------------- #
        if valid_loader is not None and (epoch % self.cfg.valid_per_epoch == 0 or self.debug_mode):
            val_log = self._valid_epoch(epoch, valid_loader)
            log.update(**{'val_'+k: v for k, v in val_log.items()})

        if self.debug_mode:
            return

        self._log_log(log)

        # --------- Evaluate model performance according to configured metric -------- #
        assert self.cfg.mnt_metric in log.keys(), '%s not in log keys' % self.cfg.mnt_metric
        self.monitor.update(log[self.cfg.mnt_metric], info={'epoch': epoch})
        self.logger.info('PeformenceMonitor: ' + str(self.monitor.state_dict()))

        # -------------------------- Save checkpoint -------------------------- #
        save_path = self._save_checkpoint(epoch, postfix='latest')
        self.logger.info("Saving latest checkpoint: {} ...".format(save_path))

        if epoch % self.cfg.save_per_epoch == 0:
            save_path = self._save_checkpoint(epoch)
            self.logger.info("Saving checkpoint: {} ...".format(save_path))

        if self.monitor.is_best():
            save_path = self._save_checkpoint(epoch, postfix='best')
            self.logger.warning("Saving current best: {} ...".format(save_path))

    def test(self, test_loader, name=None):
        old_logger = self.logger
        self.test_log_dir = self.experiment.save_dir / 'test' / 'epoch{}'.format(self.epoch-1)
        if name is not None:
            self.test_log_dir = self.test_log_dir / name
        self.test_log_dir.mkdir(parents=True, exist_ok=True)
        self.logger = Logger(self.test_log_dir)

        val_log = self._valid_epoch(1, test_loader)
        log = {'test_'+k: v for k, v in val_log.items()}

        self._log_log(log)

        self.logger = old_logger

    def _log_log(self, log):
        self._show_divider('-')
        for key, value in log.items():
            self.logger.info('    {:15s}: {}'.format(str(key), value))
        self._show_divider('-')

    def _train_epoch(self, epoch, train_loader):
        """
        Training logic for an epoch

        :param epoch: Integer, current training epoch.
        :return: A log that contains average loss and met
        ric in this epoch.
        """
        from torchvision.utils import make_grid
        
        metric_tracker = MetricTracker()

        len_epoch = len(train_loader)
        pbar = self._progress_bar(total=len_epoch)
        pbar.set_description('Train [{}|{}]'.format(epoch, self.cfg.max_epochs))
        self.module.on_epoch_start(train=True)
        for batch_idx, data in enumerate(train_loader):
            gstep = (epoch - 1) * len_epoch + batch_idx + 1
            results = self.module.step(data, train=True, epoch=epoch, step=gstep)

            self.logger.tensorboard.set_step(gstep, mode='train')
            for name, value in results.metrics.items():
                metric_tracker.update(name, value)
                self.logger.tensorboard.add_scalar(name, value)

            if gstep % self.cfg.log_step == 0:
                self.logger.debug('Train Epoch: {} {} {}'.format(epoch,
                                                                 _progress(batch_idx, train_loader),
                                                                 metric_tracker.summary()))
            if gstep % self.cfg.log_img_step == 0:
                for name, img in results.imgs.items():
                    img_name = os.path.join('train', name, '{}_{}.png'.format(epoch, gstep))
                    self.logger.save_img(img_name, make_grid(img, nrow=8, normalize=True))

            pbar.set_postfix(self._format_nums(metric_tracker.result()))
            pbar.update()

            if self.debug_mode:
                break

        pbar.close()
        self.module.on_epoch_end(train=True)

        return metric_tracker.result()

    def _valid_epoch(self, epoch, valid_loader):
        """
        Validate after training an epoch

        :param epoch: Integer, current training epoch.
        :return: A log that contains information about validation
        """
        import torch
        from torchvision.utils import make_grid
        
        metric_tracker = MetricTracker()

        len_epoch = len(valid_loader)
        pbar = self._progress_bar(total=len_epoch)
        pbar.set_description('Valid [{}|{}]'.format(epoch, self.cfg.max_epochs))
        self.module.on_epoch_start(train=False)
        with torch.no_grad():
            len_epoch = len(valid_loader)
            for batch_idx, data in enumerate(valid_loader):
                gstep = (epoch - 1) * len_epoch + batch_idx + 1
                results = self.module.step(data, train=False, epoch=epoch, step=gstep)

                self.logger.debug(f'{batch_idx}: {results.metrics}')

                self.logger.tensorboard.set_step(gstep, mode='valid')
                for name, value in results.metrics.items():
                    metric_tracker.update(name, value)
                    self.logger.tensorboard.add_scalar(name, value)

                if gstep % self.cfg.valid_log_img_step == 0:
                    for name, img in results.imgs.items():
                        img_name = os.path.join('valid', name, '{}_{}.png'.format(epoch, gstep))
                        self.logger.save_img(img_name, make_grid(img, nrow=8, normalize=True))

                pbar.set_postfix(self._format_nums(metric_tracker.result()))
                pbar.update()

                if self.debug_mode:
                    break
            pbar.close()
        self.module.on_epoch_end(train=False)
        return metric_tracker.result()

    def _save_checkpoint(self, epoch, postfix=None):
        """
        Saving checkpoints and return saved path

        :param epoch: current epoch number
        :param save_best: if True, rename the saved checkpoint to 'model-best.pth'
        """
        import torch
        
        if postfix is None:
            postfix = f'epoch{epoch}'
        state = {
            'epoch': epoch,
            'module': self.module.state_dict(),
            'monitor': self.monitor.state_dict() if self.cfg.mnt_mode != 'off' else None,
            'config': self.cfg
        }
        filename = str(self.experiment.ckpt_dir / f'model-{postfix}.pth')
        torch.save(state, filename)
        self.ckpt_cleaner.clean()
        return filename

    def _resume_checkpoint(self, resume_path):
        """
        Resume from saved checkpoints

        :param resume_path: Checkpoint path to be resumed
        """
        import torch
        
        resume_path = str(resume_path)
        self.logger.info("Loading checkpoint: {} ...".format(resume_path))
        checkpoint = torch.load(resume_path)
        self.start_epoch = checkpoint['epoch'] + 1
        self.epoch = self.start_epoch

        if 'monitor' in checkpoint:
            self.monitor.load_state_dict(checkpoint['monitor'])
        else:
            self.logger.info("Monitor state dict not found, default value will be used.")

        self.module.load_state_dict(checkpoint['module'])

        self.logger.info("Checkpoint loaded. Resume from epoch {}".format(self.start_epoch-1))

    def _progress_bar(self, total):
        if self.cfg.pbar == 'qqdm':
            bar = qqdm
        elif self.cfg.pbar == 'tqdm':
            bar = tqdm
        else:
            raise ValueError('choice of progressbar [qqdm, tqdm]')
        return bar(total=total, dynamic_ncols=True)

    def _show_divider(self, sym, text='', ncols=None):
        print(text_divider(sym, text, ncols))

    def _format_nums(self, d):
        return {k: format_num(v, fmt=self.cfg.num_fmt) for k, v in d.items()}


def _progress(batch_idx, loader):
    current = batch_idx * loader.batch_size
    total = len(loader) * loader.batch_size
    return '[{}/{} ({:.0f}%)]'.format(current, total, 100.0 * current / total)



# Created by Qixun Qu
# quqixun@gmail.com
# 2017/04/11
#


import cv2
import numpy as np
from .affine_ransac import Ransac
from .affine_transform import Affine


# The ration of the best match over second best match
#      distance of best match
# ------------------------------- <= MATCH_RATIO
#  distance of second best match
RATIO = 0.8


class Align():

    def __init__(self, K=3, threshold=1):
        ''' __INIT__

            Initialize the instance.

            Input arguments:

            - K : the number of corresponding points, default is 3
            - threshold : a threshold determins which points are outliers
            in the RANSAC process, if the residual is larger than threshold,
            it can be regarded as outliers, default value is 1

        '''

        self.K = K
        self.threshold = threshold

    def read_image(self, path, mode=1):
        ''' READ_IMAGE

            Load image from file path.

            Input arguments:

            - path : the image to be read
            - mode : 1 for reading color image, 0 for grayscale image
            default is 1

            Output:

            - the image to be processed

        '''

        return cv2.imread(path, mode)

    def extract_SIFT(self, img, max_num_kp=10000):
        ''' EXTRACT_SIFT

            Extract SIFT descriptors from the given image.

            Input argument:

            - img : the image to be processed

            Output:

            -kp : positions of key points where descriptors are extracted
            - desc : all SIFT descriptors of the image, its dimension
            will be n by 128 where n is the number of key points


        '''

        # Convert the image to grayscale
        img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        # Extract key points and SIFT descriptors
        sift = cv2.SIFT_create(max_num_kp)
        kp, desc = sift.detectAndCompute(img_gray, None)

        # Extract positions of key points
        kp = np.array([p.pt for p in kp]).T

        return kp, desc

    def match_SIFT(self, desc_s, desc_t, k=10):
        ''' MATCH_SIFT

            Match SIFT descriptors of source image and target image.
            Obtain the index of conrresponding points to do estimation
            of affine transformation.

            Input arguments:

            - desc_s : descriptors of source image
            - desc_t : descriptors of target image

            Output:

            - fit_pos : index of corresponding points

        '''

        # Match descriptor and obtain two best matches
        bf = cv2.BFMatcher()
        matches = bf.knnMatch(desc_s, desc_t, k=k)

        # Initialize output variable
        fit_pos = np.array([], dtype=np.int32).reshape((0, 2))

        matches_num = len(matches)
        for i in range(matches_num):
            # Obtain the good match if the ration id smaller than 0.8
            if matches[i][0].distance <= RATIO * matches[i][1].distance:
                temp = np.array([matches[i][0].queryIdx,
                                 matches[i][0].trainIdx])
                # Put points index of good match
                fit_pos = np.vstack((fit_pos, temp))

        return fit_pos

    def affine_matrix(self, kp_s, kp_t, fit_pos):
        ''' AFFINE_MATRIX

            Compute affine transformation matrix by corresponding points.

            Input arguments:

            - kp_s : key points from source image
            - kp_t : key points from target image
            - fit_pos : index of corresponding points

            Output:

            - M : the affine transformation matrix whose dimension
            is 2 by 3

        '''

        # Extract corresponding points from all key points
        kp_s = kp_s[:, fit_pos[:, 0]]
        kp_t = kp_t[:, fit_pos[:, 1]]

        # Apply RANSAC to find most inliers
        _, _, inliers = Ransac(self.K, self.threshold).ransac_fit(kp_s, kp_t)

        # Extract all inliers from all key points
        kp_s = kp_s[:, inliers[0]]
        kp_t = kp_t[:, inliers[0]]

        # Use all inliers to estimate transform matrix
        A, t = Affine().estimate_affine(kp_s, kp_t)
        M = np.hstack((A, t))

        return M

    def warp_image(self, source, target, M,  *args, **kwargs):
        ''' WARP_IMAGE

            Warp the source image into target with the affine
            transformation matrix.

            Input arguments:

            - source : the source image to be warped
            - target : the target image
            - M : the affine transformation matrix

            Output:

            - warp: warped source image
        '''

        # Obtain the size of target image
        rows, cols, _ = target.shape

        # Warp the source image
        warp = cv2.warpAffine(source, M, (cols, rows), *args, **kwargs)

        return warp

    def align_image(self, img_source, img_target, max_num_kp=10000, knn_k=10):
        ''' ALIGN_IMAGE

            Warp the source image into target image.
            Two images' path are provided when the
            instance Align() is created.

            Input arguments:

            - img_source(str, ndarray): the path or ndarry of sorce image that to be warped
            - img_target(str, ndarray): the path or ndarry of target image
            - max_num_kp: the max number of key points extracted by SIFT

            Output:

            - warp: warped source image

            Note: ndarray should be in cv2 format.
        '''

        # Load source image and target image
        if isinstance(img_source, str):
            img_source = self.read_image(img_source)
            img_target = self.read_image(img_target)

        # Extract key points and SIFT descriptors from
        # source image and target image respectively
        kp_s, desc_s = self.extract_SIFT(img_source, max_num_kp=max_num_kp)
        kp_t, desc_t = self.extract_SIFT(img_target, max_num_kp=max_num_kp)

        # Obtain the index of correcponding points
        fit_pos = self.match_SIFT(desc_s, desc_t, k=knn_k)

        # Compute the affine transformation matrix
        M = self.affine_matrix(kp_s, kp_t, fit_pos)

        # Warp the source image and display result
        warp = self.warp_image(img_source, img_target, M)

        return warp, M



from typing import Sequence, Optional
import random

import torch
from torch.utils.data.dataloader import DataLoader
from torch.utils.data import Dataset, Sampler
import numpy as np


# This module provide utils for controling reproducibility
# https://pytorch.org/docs/stable/notes/randomness.html


def setup_randomness(seed=1234, deterministic=True):
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)

    if deterministic:
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)


def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


def seed_generator(seed):
    g = torch.Generator()
    g.manual_seed(seed)


class SeedDataLoader(DataLoader):
    def __init__(self,
                 dataset: Dataset,
                 seed: Optional[int] = 0,
                 batch_size: Optional[int] = 1,
                 shuffle: bool = False,
                 sampler: Optional[Sampler[int]] = None,
                 batch_sampler: Optional[Sampler[Sequence[int]]] = None,
                 num_workers: int = 0,
                 collate_fn = None,
                 pin_memory: bool = False,
                 drop_last: bool = False,
                 timeout: float = 0,
                 multiprocessing_context=None, *,
                 prefetch_factor: int = 2,
                 persistent_workers: bool = False):

        super().__init__(dataset,
                         batch_size=batch_size,
                         shuffle=shuffle,
                         sampler=sampler,
                         batch_sampler=batch_sampler,
                         num_workers=num_workers,
                         collate_fn=collate_fn,
                         pin_memory=pin_memory,
                         drop_last=drop_last,
                         timeout=timeout,
                         worker_init_fn=seed_worker,  # use seed worker
                         multiprocessing_context=multiprocessing_context,
                         generator=seed_generator(seed),  # use seed generator
                         prefetch_factor=prefetch_factor,
                         persistent_workers=persistent_workers)



import json
from pathlib import Path
from collections import OrderedDict
from functools import partial
import os
from operator import attrgetter
from typing import Sequence

import torch


def get_obj(info, module, *args, **kwargs):
    """
    Finds a function handle with the name given as 'type@' in info, and returns the
    instance initialized with corresponding arguments given.

    `object = get_obj(info['type@'], module)`
    is equivalent to
    `object = module.info['type@'](info.pop('type@'))`
    """
    module_name = info['type@']
    module_args = dict(info)
    module_args.pop('type@')
    assert all([k not in module_args for k in kwargs]), 'Overwriting kwargs given in config file is not allowed'
    module_args.update(kwargs)
    return attrgetter(module_name)(module)(*args, **module_args)


def get_ftn(info, module, *args, **kwargs):
    """
    Finds a function handle with the name given as 'type' in info, and returns the
    function with given arguments fixed with functools.partial.

    `function = get_ftn('name', module, a, b=1)`
    is equivalent to
    `function = lambda *args, **kwargs: module.name(a, *args, b=1, **kwargs)`.
    """
    module_name = info['type']
    module_args = dict(info['args'])
    assert all([k not in module_args for k in kwargs]), 'Overwriting kwargs given in config file is not allowed'
    module_args.update(kwargs)
    return partial(getattr(module, module_name), *args, **module_args)



def adjust_learning_rate(optimizer, lr):
    print('Adjust Learning Rate => %.4e' % lr)
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr
        # param_group['initial_lr'] = lr


def ensure_dir(dirname):
    dirname = Path(dirname)
    if not dirname.is_dir():
        dirname.mkdir(parents=True, exist_ok=False)


def read_json(fname):
    fname = Path(fname)
    with fname.open('rt') as handle:
        return json.load(handle, object_hook=OrderedDict)


def write_json(content, fname):
    fname = Path(fname)
    with fname.open('wt') as handle:
        json.dump(content, handle, indent=4, sort_keys=False)


def prepare_device(n_gpu_use):
    """
    setup GPU device if available. get gpu device indices which are used for DataParallel
    """
    n_gpu = torch.cuda.device_count()
    if n_gpu_use > 0 and n_gpu == 0:
        print("Warning: There\'s no GPU available on this machine,"
              "training will be performed on CPU.")
        n_gpu_use = 0
    if n_gpu_use > n_gpu:
        print(f"Warning: The number of GPU\'s configured to use is {n_gpu_use}, but only {n_gpu} are "
              "available on this machine.")
        n_gpu_use = n_gpu
    device = torch.device('cuda:0' if n_gpu_use > 0 else 'cpu')
    list_ids = list(range(n_gpu_use))
    return device, list_ids


def to_device(data, device):
    if isinstance(data, torch.Tensor):
        return data.to(device).float()
    if isinstance(data, Sequence):
        return [to_device(d, device=device) for d in data]
    if isinstance(data, dict):
        return {k: to_device(v, device=device) for k, v in data.items()}
    return data


def load_checkpoint(model, ckpt_path):
    model.load_state_dict(torch.load(ckpt_path)['module']['model'])


def auto_rename(path):
    count = 1
    new_path = path
    while True:
        if not os.path.exists(new_path):
            return new_path
        file_name = os.path.basename(path)
        name, ext = file_name.split('.')
        new_file_name = '{}_{}.{}'.format(name, count, ext)
        new_path = os.path.join(os.path.dirname(path), new_file_name)
        count += 1


class ConfigurableLossCalculator:
    def __init__(self, cfg: dict):
        self.cfg = cfg
        self.loss_fns = {}

    def register(self, fn, name):
        self.loss_fns[name] = fn

    def compute(self):
        loss = 0
        loss_dict = {}
        for name, weight in self.cfg.items():
            l = self.loss_fns[name]() * weight
            loss_dict[name] = l.item()
            loss += l
        return loss, loss_dict



import os

import numpy as np
from scipy import ndimage
import scipy
import cv2
from scipy.io import loadmat

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def fspecial_gaussian(hsize, sigma):
    hsize = [hsize, hsize]
    siz = [(hsize[0]-1.0)/2.0, (hsize[1]-1.0)/2.0]
    std = sigma
    [x, y] = np.meshgrid(np.arange(-siz[1], siz[1]+1),
                         np.arange(-siz[0], siz[0]+1))
    arg = -(x*x + y*y)/(2*std*std)
    h = np.exp(arg)
    h[h < scipy.finfo(float).eps * h.max()] = 0
    sumh = h.sum()
    if sumh != 0:
        h = h/sumh
    return h


class AbstractBlur:
    def __call__(self, img):
        img_L = ndimage.filters.convolve(
            img, np.expand_dims(self.kernel, axis=2), mode='nearest')
        return img_L


class GaussianBlur(AbstractBlur):
    """ Expect input'shape = [H,W,C]
    """

    def __init__(self, ksize=8, sigma=3):
        self.kernel = fspecial_gaussian(ksize, sigma)


class UniformBlur(AbstractBlur):
    """ Expect input'shape = [H,W,C]
    """

    def __init__(self, ksize):
        self.kernel = np.ones((ksize, ksize)) / (ksize*ksize)


## -------------------- Downsample -------------------- ##

class KFoldDownsample:
    ''' k-fold downsampler:
        Keeping the upper-left pixel for each distinct sfxsf patch and discarding the others

        Expect input'shape = [H,W,C]
    '''

    def __init__(self, sf):
        self.sf = sf

    def __call__(self, img):
        """ input: [H,W,c] """
        st = 0
        return img[st::self.sf, st::self.sf, :]


class AbstractDownsample:
    def __call__(self, img):
        img = self.blur(img)
        img = self.downsampler(img)
        return img


class UniformDownsample(AbstractDownsample):
    """ Expect input'shape = [H,W,C]
    """

    def __init__(self, sf):
        self.sf = sf
        self.blur = UniformBlur(sf)
        self.downsampler = KFoldDownsample(sf)


class GaussianDownsample(AbstractDownsample):
    """ Expect input'shape = [H,W,C]
    """

    def __init__(self, sf, ksize=8, sigma=3):
        self.sf = sf
        self.blur = GaussianBlur(ksize, sigma)
        self.downsampler = KFoldDownsample(sf)



class Resize:
    """ Expect input'shape = [H,W,C]
    """

    def __init__(self, sf, mode='cubic'):
        self.sf = sf
        self.mode = self.mode_map[mode]

    def __call__(self, img):
        return cv2.resize(img, (int(img.shape[1]*self.sf), int(img.shape[0]*self.sf)), interpolation=self.mode)

    mode_map = {
        'nearest': cv2.INTER_NEAREST,
        'linear': cv2.INTER_LINEAR,
        'cubic': cv2.INTER_CUBIC,
        'area': cv2.INTER_AREA
    }


Upsample = Resize # backward capability

class HSI2RGB(object):
    def __init__(self, spe=None):
        if spe is None:
            CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
            self.SPE = loadmat(os.path.join(CURRENT_DIR, 'kernels', 'misr_spe_p.mat'))['P']  # (3,31)
        else:
            self.SPE = spe

    def __call__(self, img):
        return img @ self.SPE.transpose()



import numpy as np

from .functional import hflip, vflip, brightness_change, multiplicative_color_change, contrast_change, rotate


# assume channel first

class Augment:
    def __call__(self, img):
        raise NotImplementedError


class Sequential(Augment):
    def __init__(self, augments):
        self.augments = augments

    def __call__(self, img):
        for augment in self.augments:
            img = augment(img)
        return img


class FlipHorizontal(Augment):
    def __init__(self):
        self.enable = np.random.randint(0, 2) > 0.5

    def __call__(self, img):
        if self.enable:
            return hflip(img)
        return img


class FlipVertical(Augment):
    def __init__(self):
        self.enable = np.random.randint(0, 2) > 0.5

    def __call__(self, img):
        if self.enable:
            return vflip(img)
        return img


class BrightnessChange(Augment):
    def __init__(self):
        self.strength = np.random.normal(loc=0, scale=0.02)

    def __call__(self, img):
        return brightness_change(img, self.strength)


class MultiplicativeColorChange(Augment):
    def __init__(self):
        self.strength = np.random.uniform(0.9, 1.1)

    def __call__(self, img):
        return multiplicative_color_change(img, self.strength)


class Contrast(Augment):
    def __init__(self):
        self.alpha = np.random.uniform(-0.3, 0.3)

    def __call__(self, img):
        return contrast_change(img, self.alpha)


# TODO: Fix
class Rotate(Augment):
    def __init__(self):
        self.k = np.random.randint(0, 4)

    def __call__(self, img):
        return rotate(img, self.k)





# Created by Qixun Qu
# quqixun@gmail.com
# 2017/04/10
#


import numpy as np


class Affine():

    def create_test_case(self, outlier_rate=0):
        ''' CREATE_TEST_CASE

            Randomly generate a test case of affine transformation.

            Input arguments:

            - outlier_rate : the percentage of outliers in test case,
            default is 0

            Outputs:

            - pts : warped points
            - pts_tilde : source points that wll be transformed
            - A, t : parameters of affine transformation, A is a 2x2
            matrix, t is a 2x1 vector, both of them are created randomly

        '''

        # Randomly generate affine transformation
        # A is a 2x2 matrix, the range of each value is from -2 to 2
        A = 4 * np.random.rand(2, 2) - 2

        # % t is a 2x1 VECTOR, the range of each value is from -10 to 10
        t = 20 * np.random.rand(2, 1) - 10

        # Set the number of points in test case
        num = 1000

        # Compute the number of outliers and inliers respectively
        outliers = int(np.round(num * outlier_rate))
        inliers = int(num - outliers)

        # Gernerate source points whose scope from (0,0) to (100, 100)
        pts_s = 100 * np.random.rand(2, num)
        # Initialize warped points matrix
        pts_t = np.zeros((2, num))

        # Compute inliers in warped points matrix by applying A and t
        pts_t[:, :inliers] = np.dot(A, pts_s[:, :inliers]) + t

        # Generate outliers in warped points matrix
        pts_t[:, inliers:] = 100 * np.random.rand(2, outliers)

        # Reset the order of warped points matrix,
        # outliers and inliers will scatter randomly in test case
        rnd_idx = np.random.permutation(num)
        pts_s = pts_s[:, rnd_idx]
        pts_t = pts_t[:, rnd_idx]

        return A, t, pts_s, pts_t

    def estimate_affine(self, pts_s, pts_t):
        ''' ESTIMATE_AFFINE

            Estimate affine transformation by the given points
            correspondences.

            Input arguments:
            - pts : points in target image
            - pts_tilde : points in source image

            Outputs:

            - A, t : the affine transformation, A is a 2x2 matrix
            that indicates the rotation and scaling transformation,
            t is a 2x1 vector determines the translation

            Method:

            To estimate an affine transformation between two images,
            at least 3 corresponding points are needed.
            In this case, 6-parameter affine transformation are taken into
            consideration, which is shown as follows:

            | x' | = | a b | * | x | + | tx |
            | y' |   | c d |   | y |   | ty |

            For 3 corresponding points, 6 equations can be formed as below:

            | x1 y1 0  0  1 0 |       | a  |       | x1' |
            | 0  0  x1 y1 0 1 |       | b  |       | y1' |
            | x2 y2 0  0  1 0 |   *   | c  |   =   | x2' |
            | 0  0  x2 y2 0 1 |       | d  |       | y2' |
            | x3 y3 0  0  1 0 |       | tx |       | x3' |
            | 0  0  x3 y3 0 1 |       | ty |       | y3' |

            |------> M <------|   |-> theta <-|   |-> b <-|

            Solve the equation to compute theta by:  theta = M \ b
            Thus, affine transformation can be obtained as:

            A = | a b |     t = | tx |
                | c d |         | ty |

        '''

        # Get the number of corresponding points
        pts_num = pts_s.shape[1]

        # Initialize the matrix M,
        # M has 6 columns, since the affine transformation
        # has 6 parameters in this case
        M = np.zeros((2 * pts_num, 6))

        for i in range(pts_num):
            # Form the matrix m
            temp = [[pts_s[0, i], pts_s[1, i], 0, 0, 1, 0],
                    [0, 0, pts_s[0, i], pts_s[1, i], 0, 1]]
            M[2 * i: 2 * i + 2, :] = np.array(temp, dtype=object)

        # Form the matrix b,
        # b contains all known target points
        b = pts_t.T.reshape((2 * pts_num, 1))

        try:
            # Solve the linear equation
            theta = np.linalg.lstsq(M, b, rcond=None)[0]

            # Form the affine transformation
            A = theta[:4].reshape((2, 2))
            t = theta[4:]
        except np.linalg.linalg.LinAlgError:
            # If M is singular matrix, return None
            # print("Singular matrix.")
            A = None
            t = None

        return A, t



import numpy as np
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
from torch.utils.data.sampler import SubsetRandomSampler
from torchvision import datasets, transforms


class BaseDataLoader(DataLoader):
    """
    Base class for all data loaders
    """

    def __init__(
            self, dataset, batch_size, shuffle=True, validation_split=0, num_workers=4, collate_fn=default_collate, **
            kwargs):
        self.validation_split = validation_split
        self.shuffle = shuffle

        self.batch_idx = 0
        self.n_samples = len(dataset)

        self.sampler, self.valid_sampler = self._split_sampler(self.validation_split)

        self.init_kwargs = kwargs
        self.init_kwargs.update({
            'dataset': dataset,
            'batch_size': batch_size,
            'shuffle': self.shuffle,
            'collate_fn': collate_fn,
            'num_workers': num_workers
        })
        super().__init__(sampler=self.sampler, **self.init_kwargs)

    def _split_sampler(self, split):
        if split == 0.0:
            return None, None

        idx_full = np.arange(self.n_samples)

        np.random.seed(0)
        np.random.shuffle(idx_full)

        if isinstance(split, int):
            assert split > 0
            assert split < self.n_samples, "validation set size is configured to be larger than entire dataset."
            len_valid = split
        else:
            len_valid = int(self.n_samples * split)

        valid_idx = idx_full[0:len_valid]
        train_idx = np.delete(idx_full, np.arange(0, len_valid))

        train_sampler = SubsetRandomSampler(train_idx)
        valid_sampler = SubsetRandomSampler(valid_idx)

        # turn off shuffle option which is mutually exclusive with sampler
        self.shuffle = False
        self.n_samples = len(train_idx)

        return train_sampler, valid_sampler

    def split_validation(self):
        if self.valid_sampler is None:
            return None
        else:
            return DataLoader(sampler=self.valid_sampler, **self.init_kwargs)



import numpy as np
import random


def minmax_normalize(array):
    amin = np.min(array)
    amax = np.max(array)
    return (array - amin) / (amax - amin)


def crop_center(img, cropx, cropy):
    x, y = img.shape[-2], img.shape[-1]
    startx = x//2-(cropx//2)
    starty = y//2-(cropy//2)
    return img[..., startx:startx+cropx, starty:starty+cropy]


def rand_crop(img, cropx, cropy):
    x, y = img.shape[-2], img.shape[-1]
    x1 = random.randint(0, x - cropx)
    y1 = random.randint(0, y - cropy)
    return img[..., x1:x1+cropx, y1:y1+cropy]


def mod_crop(img, modulo):
    _, ih, iw = img.shape
    ih = ih - (ih % modulo)
    iw = iw - (iw % modulo)
    img = img[:, 0:ih, 0:iw]
    return img



import numpy as np

def vflip(img):
    return img[..., ::-1, :]


def hflip(img):
    return img[..., :, ::-1]


def brightness_change(img, strength):
    return img + strength


def multiplicative_color_change(img, strength):
    return img * strength


def grayscale(img):
    dst = np.zeros((1, img.shape[1], img.shape[2]), dtype=np.float32)
    dst[0, :, :] = 0.299 * img[0, :, :] + 0.587 * img[1, :, :] + 0.114 * img[2, :, :]
    dst = np.repeat(dst, 3, axis=0)
    return dst


def blend(img1, img2, alpha=0.5):
    return img1 * alpha + img2 * (1-alpha)


def contrast_change(img, alpha):
    gs_2 = grayscale(img)
    img = blend(gs_2, img, alpha=alpha)
    return img


def rotate(img, k):
    return np.rot90(img, k)




from abc import ABC, abstractmethod
from typing import NamedTuple


class Module(ABC):
    class StepResult(NamedTuple):
        imgs: dict = {}
        metrics: dict = {}

    @abstractmethod
    def step(self, data, train, epoch, step) -> StepResult:
        """ return a StepResult that contains the imgs and metrics you want to save and log """
        raise NotImplementedError

    def on_epoch_start(self, train):
        pass

    def on_epoch_end(self, train):
        pass

    @abstractmethod
    def state_dict(self):
        raise NotImplementedError

    @abstractmethod
    def load_state_dict(self, state):
        raise NotImplementedError


class SMSOModule(Module):
    """ Single Model Single Optimizer Module"""

    def __init__(self, model, optimizer):
        super().__init__()
        self.model = model
        self.optimizer = optimizer

    def step(self, data, train, epoch, step):
        if train:
            self.model.train()
            self.optimizer.zero_grad()

        loss, result = self._step(data, train, epoch, step)

        if train:
            loss.backward()
            self.optimizer.step()

        return result

    def _step(self, data, train, epoch, step):
        """ return (loss, self.StepResult) 
        """
        raise NotImplementedError

    def state_dict(self):
        return {'model': self.model.state_dict(),
                'optimizer': self.optimizer.state_dict()}

    def load_state_dict(self, state):
        self.model.load_state_dict(state['model'])
        if 'optimizer' in state:
            self.optimizer.load_state_dict(state['optimizer'])


class SimpleModule(SMSOModule):
    def __init__(self, model, optimizer, device):
        super().__init__(model, optimizer)
        self.device = device
        self.model = self.model.to(device)

    def _step(self, data, train, epoch, step):
        input, target = data
        input, target = input.to(self.device), target.to(self.device)
        output = self.model(input)
        loss = self.criterion(output, target)

        metrics = {'loss': loss.item()}
        return loss, self.StepResult(metrics=metrics)



import logging
import logging.config
import os
from pathlib import Path
import os
import yaml
import importlib 
from datetime import datetime


CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def setup_logging(save_dir, log_config=os.path.join(CURRENT_DIR, 'logger_config.yaml'), default_level=logging.INFO):
    """
    Setup logging configuration
    """
    log_config = Path(log_config)
    save_dir = Path(save_dir)
    save_dir.mkdir(exist_ok=True)
    if log_config.is_file():
        with log_config.open('rt') as handle:
            config = yaml.load(handle, Loader=yaml.FullLoader)
        # modify logging paths based on run config
        for _, handler in config['handlers'].items():
            if 'filename' in handler:
                handler['filename'] = str(save_dir / handler['filename'])

        logging.config.dictConfig(config)
    else:
        print("Warning: logging configuration file is not found in {}.".format(log_config))
        logging.basicConfig(level=default_level)


log_levels = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG
}


def get_logger(name, save_dir, verbosity=2):
    import colorlog
    setup_logging(save_dir)
    msg_verbosity = 'verbosity option {} is invalid. Valid options are {}.'.format(
        verbosity, log_levels.keys())
    assert verbosity in log_levels, msg_verbosity
    logger = colorlog.getLogger(name)
    logger.setLevel(log_levels[verbosity])
    return logger


class Logger:
    def __init__(self, log_dir, enable_tensorboard=False):
        self.log_dir = Path(log_dir)
        self.tensorboard_ = None
        self.enable_tensorboard = enable_tensorboard
        self.text = get_logger('Torchlight', self.log_dir)
        self.img_dir = self.log_dir / 'img'

    @property
    def tensorboard(self):
        if self.tensorboard_ is None:
            tensorboard_dir = os.path.join(self.log_dir, 'tensorboard')
            self.tensorboard_ = TensorboardWriter(tensorboard_dir, logger=self, 
                                                  enabled=self.enable_tensorboard)
        return self.tensorboard_

    def info(self, msg):
        self.text.info(msg)

    def debug(self, msg):
        self.text.debug(msg)

    def warning(self, msg):
        self.text.warning(msg)

    def save_img(self, name, img):
        # TODO: remove torchvision dependency
        from torchvision.utils import save_image
        save_path: Path = self.img_dir / name
        if not save_path.parent.exists():
            save_path.parent.mkdir(parents=True)
        save_image(img, save_path)


class TensorboardWriter():
    def __init__(self, log_dir, logger, enabled):
        self.writer = None
        self.selected_module = ""

        if enabled:
            log_dir = str(log_dir)

            # Retrieve vizualization writer.
            succeeded = False
            for module in ["torch.utils.tensorboard", "tensorboardX"]:
                try:
                    self.writer = importlib.import_module(module).SummaryWriter(log_dir)
                    succeeded = True
                    break
                except ImportError:
                    succeeded = False
                self.selected_module = module

            if not succeeded:
                message = "Warning: visualization (Tensorboard) is configured to use, but currently not installed on " \
                    "this machine. Please install TensorboardX with 'pip install tensorboardx', upgrade PyTorch to " \
                    "version >= 1.1 to use 'torch.utils.tensorboard' or turn off the option in the 'config.json' file."
                logger.warning(message)

        self.step = 0
        self.mode = ''

        self.tb_writer_ftns = {
            'add_scalar', 'add_scalars', 'add_image', 'add_images', 'add_audio',
            'add_text', 'add_histogram', 'add_pr_curve', 'add_embedding'
        }
        self.tag_mode_exceptions = {'add_histogram', 'add_embedding'}
        self.timer = datetime.now()

    def set_step(self, step, mode='train'):
        self.mode = mode
        self.step = step
        if step == 0:
            self.timer = datetime.now()
        else:
            duration = datetime.now() - self.timer
            self.add_scalar('steps_per_sec', 1 / duration.total_seconds())
            self.timer = datetime.now()

    def __getattr__(self, name):
        """
        If visualization is configured to use:
            return add_data() methods of tensorboard with additional information (step, tag) added.
        Otherwise:
            return a blank function handle that does nothing
        """
        if name in self.tb_writer_ftns:
            # if enable is false, writer is None, then add_data would be none
            add_data = getattr(self.writer, name, None)

            def wrapper(tag, data, *args, **kwargs):
                if add_data is not None:
                    # add mode(train/valid) tag
                    if name not in self.tag_mode_exceptions:
                        tag = '{}/{}'.format(tag, self.mode)
                    add_data(tag, data, self.step, *args, **kwargs)
            return wrapper
        else:
            # default action for returning methods defined in this class, set_step() for instance.
            try:
                attr = object.__getattr__(name)
            except AttributeError:
                raise AttributeError("type object '{}' has no attribute '{}'".format(self.selected_module, name))
            return attr


import os
from typing import Union

from torch.utils.data import DataLoader

from .module import Module
from .engine import Engine
from .config import write_yaml
from ._util import auto_rename


class DataSource:
    def train_loader(self):
        raise NotImplementedError

    def valid_loader(self):
        return self.test_loader()

    def test_loader(self):
        return None


def run_lazy(args, cfg, module: Module, data_source: DataSource):
    engine = Engine(module, save_dir=args.save_dir, **cfg['engine'])

    if args.mode == 'train':
        cfg['engine'].update(engine.cfg._asdict())
        write_yaml(cfg, os.path.join(args.save_dir, 'config.yaml'))  # override the default config.yaml
        write_yaml(cfg, auto_rename(os.path.join(args.save_dir, 'history', 'config.yaml')))  # save a copy

        if args.resume:
            engine.resume(args.resume, base_dir=args.resume_dir)
        engine.train(data_source.train_loader(), valid_loader=data_source.valid_loader())
    elif args.mode == 'test':
        test_loader = data_source.test_loader()
        assert test_loader is not None, 'test loader is not provided'
        engine.resume(args.resume)

        if isinstance(test_loader, dict):
            for name, dataloader in test_loader.items():
                engine.logger.info('Start testing {} ...'.format(name))
                engine.test(dataloader, name)
        else:
            engine.test(test_loader)

        cfg['engine'].update(engine.cfg._asdict())
        write_yaml(cfg, os.path.join(engine.test_log_dir, 'config.yaml'))
    else:
        engine.set_debug_mode()
        engine.train(data_source.train_loader(), valid_loader=data_source.valid_loader())



def run(args, cfg,
        module: Module,
        train_loader: DataLoader,
        valid_loader: DataLoader = None,
        test_loader: Union[DataLoader, dict] = None):
    """ test_loader can be a dict of Dataloaders in the format of {name: loader}
        ```python
        run(args, cfg, module, train_loader, valid_loader, test_loader)
        run(args, cfg, module, train_loader, valid_loader, {'a': test_loader1})
        ``` 
    """

    engine = Engine(module, save_dir=args.save_dir, **cfg['engine'])

    if args.mode == 'train':
        cfg['engine'].update(engine.cfg._asdict())
        write_yaml(cfg, os.path.join(args.save_dir, 'config.yaml'))  # override the default config.yaml
        write_yaml(cfg, auto_rename(os.path.join(args.save_dir, 'history', 'config.yaml')))  # save a copy

        if args.resume:
            engine.resume(args.resume, base_dir=args.resume_dir)
        engine.train(train_loader, valid_loader=valid_loader)
    elif args.mode == 'test':
        assert test_loader is not None, 'test loader is not provided'
        engine.resume(args.resume)

        if isinstance(test_loader, dict):
            for name, dataloader in test_loader.items():
                engine.logger.info('Start testing {} ...'.format(name))
                engine.test(dataloader, name)
        else:
            engine.test(test_loader)

        cfg['engine'].update(engine.cfg._asdict())
        write_yaml(cfg, os.path.join(engine.test_log_dir, 'config.yaml'))
    else:
        engine.set_debug_mode()
        engine.train(train_loader, valid_loader=valid_loader)



import math
import numpy as np
import torch


def cubic(x):
    """cubic function used for calculate_weights_indices."""
    absx = torch.abs(x)
    absx2 = absx**2
    absx3 = absx**3
    return (1.5 * absx3 - 2.5 * absx2 + 1) * (
        (absx <= 1).type_as(absx)) + (-0.5 * absx3 + 2.5 * absx2 - 4 * absx + 2) * (((absx > 1) *
                                                                                     (absx <= 2)).type_as(absx))


def calculate_weights_indices(in_length, out_length, scale, kernel, kernel_width, antialiasing):
    """Calculate weights and indices, used for imresize function.
    Args:
        in_length (int): Input length.
        out_length (int): Output length.
        scale (float): Scale factor.
        kernel_width (int): Kernel width.
        antialisaing (bool): Whether to apply anti-aliasing when downsampling.
    """

    if (scale < 1) and antialiasing:
        # Use a modified kernel (larger kernel width) to simultaneously
        # interpolate and antialias
        kernel_width = kernel_width / scale

    # Output-space coordinates
    x = torch.linspace(1, out_length, out_length)

    # Input-space coordinates. Calculate the inverse mapping such that 0.5
    # in output space maps to 0.5 in input space, and 0.5 + scale in output
    # space maps to 1.5 in input space.
    u = x / scale + 0.5 * (1 - 1 / scale)

    # What is the left-most pixel that can be involved in the computation?
    left = torch.floor(u - kernel_width / 2)

    # What is the maximum number of pixels that can be involved in the
    # computation?  Note: it's OK to use an extra pixel here; if the
    # corresponding weights are all zero, it will be eliminated at the end
    # of this function.
    p = math.ceil(kernel_width) + 2

    # The indices of the input pixels involved in computing the k-th output
    # pixel are in row k of the indices matrix.
    indices = left.view(out_length, 1).expand(out_length, p) + torch.linspace(0, p - 1, p).view(1, p).expand(
        out_length, p)

    # The weights used to compute the k-th output pixel are in row k of the
    # weights matrix.
    distance_to_center = u.view(out_length, 1).expand(out_length, p) - indices

    # apply cubic kernel
    if (scale < 1) and antialiasing:
        weights = scale * cubic(distance_to_center * scale)
    else:
        weights = cubic(distance_to_center)

    # Normalize the weights matrix so that each row sums to 1.
    weights_sum = torch.sum(weights, 1).view(out_length, 1)
    weights = weights / weights_sum.expand(out_length, p)

    # If a column in weights is all zero, get rid of it. only consider the
    # first and last column.
    weights_zero_tmp = torch.sum((weights == 0), 0)
    if not math.isclose(weights_zero_tmp[0], 0, rel_tol=1e-6):
        indices = indices.narrow(1, 1, p - 2)
        weights = weights.narrow(1, 1, p - 2)
    if not math.isclose(weights_zero_tmp[-1], 0, rel_tol=1e-6):
        indices = indices.narrow(1, 0, p - 2)
        weights = weights.narrow(1, 0, p - 2)
    weights = weights.contiguous()
    indices = indices.contiguous()
    sym_len_s = -indices.min() + 1
    sym_len_e = indices.max() - in_length
    indices = indices + sym_len_s - 1
    return weights, indices, int(sym_len_s), int(sym_len_e)


@torch.no_grad()
def imresize(img, scale, antialiasing=True):
    """imresize function same as MATLAB.
    It now only supports bicubic.
    The same scale applies for both height and width.
    Args:
        img (Tensor | Numpy array):
            Tensor: Input image with shape (c, h, w), [0, 1] range.
            Numpy: Input image with shape (h, w, c), [0, 1] range.
        scale (float): Scale factor. The same scale applies for both height
            and width.
        antialisaing (bool): Whether to apply anti-aliasing when downsampling.
            Default: True.
    Returns:
        Tensor: Output image with shape (c, h, w), [0, 1] range, w/o round.
    """
    squeeze_flag = False
    if type(img).__module__ == np.__name__:  # numpy type
        numpy_type = True
        if img.ndim == 2:
            img = img[:, :, None]
            squeeze_flag = True
        img = torch.from_numpy(img.transpose(2, 0, 1)).float()
    else:
        numpy_type = False
        if img.ndim == 2:
            img = img.unsqueeze(0)
            squeeze_flag = True

    in_c, in_h, in_w = img.size()
    out_h, out_w = math.ceil(in_h * scale), math.ceil(in_w * scale)
    kernel_width = 4
    kernel = 'cubic'

    # get weights and indices
    weights_h, indices_h, sym_len_hs, sym_len_he = calculate_weights_indices(in_h, out_h, scale, kernel, kernel_width,
                                                                             antialiasing)
    weights_w, indices_w, sym_len_ws, sym_len_we = calculate_weights_indices(in_w, out_w, scale, kernel, kernel_width,
                                                                             antialiasing)
    # process H dimension
    # symmetric copying
    img_aug = torch.FloatTensor(in_c, in_h + sym_len_hs + sym_len_he, in_w)
    img_aug.narrow(1, sym_len_hs, in_h).copy_(img)

    sym_patch = img[:, :sym_len_hs, :]
    inv_idx = torch.arange(sym_patch.size(1) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(1, inv_idx)
    img_aug.narrow(1, 0, sym_len_hs).copy_(sym_patch_inv)

    sym_patch = img[:, -sym_len_he:, :]
    inv_idx = torch.arange(sym_patch.size(1) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(1, inv_idx)
    img_aug.narrow(1, sym_len_hs + in_h, sym_len_he).copy_(sym_patch_inv)

    out_1 = torch.FloatTensor(in_c, out_h, in_w)
    kernel_width = weights_h.size(1)
    for i in range(out_h):
        idx = int(indices_h[i][0])
        for j in range(in_c):
            out_1[j, i, :] = img_aug[j, idx:idx + kernel_width, :].transpose(0, 1).mv(weights_h[i])

    # process W dimension
    # symmetric copying
    out_1_aug = torch.FloatTensor(in_c, out_h, in_w + sym_len_ws + sym_len_we)
    out_1_aug.narrow(2, sym_len_ws, in_w).copy_(out_1)

    sym_patch = out_1[:, :, :sym_len_ws]
    inv_idx = torch.arange(sym_patch.size(2) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(2, inv_idx)
    out_1_aug.narrow(2, 0, sym_len_ws).copy_(sym_patch_inv)

    sym_patch = out_1[:, :, -sym_len_we:]
    inv_idx = torch.arange(sym_patch.size(2) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(2, inv_idx)
    out_1_aug.narrow(2, sym_len_ws + in_w, sym_len_we).copy_(sym_patch_inv)

    out_2 = torch.FloatTensor(in_c, out_h, out_w)
    kernel_width = weights_w.size(1)
    for i in range(out_w):
        idx = int(indices_w[i][0])
        for j in range(in_c):
            out_2[j, :, i] = out_1_aug[j, :, idx:idx + kernel_width].mv(weights_w[i])

    if squeeze_flag:
        out_2 = out_2.squeeze(0)
    if numpy_type:
        out_2 = out_2.numpy()
        if not squeeze_flag:
            out_2 = out_2.transpose(1, 2, 0)

    return out_2


def rgb2ycbcr(img, y_only=False):
    """Convert a RGB image to YCbCr image.
    This function produces the same results as Matlab's `rgb2ycbcr` function.
    It implements the ITU-R BT.601 conversion for standard-definition
    television. See more details in
    https://en.wikipedia.org/wiki/YCbCr#ITU-R_BT.601_conversion.
    It differs from a similar function in cv2.cvtColor: `RGB <-> YCrCb`.
    In OpenCV, it implements a JPEG conversion. See more details in
    https://en.wikipedia.org/wiki/YCbCr#JPEG_conversion.
    Args:
        img (ndarray): The input image. It accepts:
            1. np.uint8 type with range [0, 255];
            2. np.float32 type with range [0, 1].
        y_only (bool): Whether to only return Y channel. Default: False.
    Returns:
        ndarray: The converted YCbCr image. The output image has the same type
            and range as input image.
    """
    img_type = img.dtype
    img = _convert_input_type_range(img)
    if y_only:
        out_img = np.dot(img, [65.481, 128.553, 24.966]) + 16.0
    else:
        out_img = np.matmul(
            img, [[65.481, -37.797, 112.0], [128.553, -74.203, -93.786], [24.966, 112.0, -18.214]]) + [16, 128, 128]
    out_img = _convert_output_type_range(out_img, img_type)
    return out_img


def bgr2ycbcr(img, y_only=False):
    """Convert a BGR image to YCbCr image.
    The bgr version of rgb2ycbcr.
    It implements the ITU-R BT.601 conversion for standard-definition
    television. See more details in
    https://en.wikipedia.org/wiki/YCbCr#ITU-R_BT.601_conversion.
    It differs from a similar function in cv2.cvtColor: `BGR <-> YCrCb`.
    In OpenCV, it implements a JPEG conversion. See more details in
    https://en.wikipedia.org/wiki/YCbCr#JPEG_conversion.
    Args:
        img (ndarray): The input image. It accepts:
            1. np.uint8 type with range [0, 255];
            2. np.float32 type with range [0, 1].
        y_only (bool): Whether to only return Y channel. Default: False.
    Returns:
        ndarray: The converted YCbCr image. The output image has the same type
            and range as input image.
    """
    img_type = img.dtype
    img = _convert_input_type_range(img)
    if y_only:
        out_img = np.dot(img, [24.966, 128.553, 65.481]) + 16.0
    else:
        out_img = np.matmul(
            img, [[24.966, 112.0, -18.214], [128.553, -74.203, -93.786], [65.481, -37.797, 112.0]]) + [16, 128, 128]
    out_img = _convert_output_type_range(out_img, img_type)
    return out_img


def ycbcr2rgb(img):
    """Convert a YCbCr image to RGB image.
    This function produces the same results as Matlab's ycbcr2rgb function.
    It implements the ITU-R BT.601 conversion for standard-definition
    television. See more details in
    https://en.wikipedia.org/wiki/YCbCr#ITU-R_BT.601_conversion.
    It differs from a similar function in cv2.cvtColor: `YCrCb <-> RGB`.
    In OpenCV, it implements a JPEG conversion. See more details in
    https://en.wikipedia.org/wiki/YCbCr#JPEG_conversion.
    Args:
        img (ndarray): The input image. It accepts:
            1. np.uint8 type with range [0, 255];
            2. np.float32 type with range [0, 1].
    Returns:
        ndarray: The converted RGB image. The output image has the same type
            and range as input image.
    """
    img_type = img.dtype
    img = _convert_input_type_range(img) * 255
    out_img = np.matmul(img, [[0.00456621, 0.00456621, 0.00456621], [0, -0.00153632, 0.00791071],
                              [0.00625893, -0.00318811, 0]]) * 255.0 + [-222.921, 135.576, -276.836]  # noqa: E126
    out_img = _convert_output_type_range(out_img, img_type)
    return out_img


def ycbcr2bgr(img):
    """Convert a YCbCr image to BGR image.
    The bgr version of ycbcr2rgb.
    It implements the ITU-R BT.601 conversion for standard-definition
    television. See more details in
    https://en.wikipedia.org/wiki/YCbCr#ITU-R_BT.601_conversion.
    It differs from a similar function in cv2.cvtColor: `YCrCb <-> BGR`.
    In OpenCV, it implements a JPEG conversion. See more details in
    https://en.wikipedia.org/wiki/YCbCr#JPEG_conversion.
    Args:
        img (ndarray): The input image. It accepts:
            1. np.uint8 type with range [0, 255];
            2. np.float32 type with range [0, 1].
    Returns:
        ndarray: The converted BGR image. The output image has the same type
            and range as input image.
    """
    img_type = img.dtype
    img = _convert_input_type_range(img) * 255
    out_img = np.matmul(img, [[0.00456621, 0.00456621, 0.00456621], [0.00791071, -0.00153632, 0],
                              [0, -0.00318811, 0.00625893]]) * 255.0 + [-276.836, 135.576, -222.921]  # noqa: E126
    out_img = _convert_output_type_range(out_img, img_type)
    return out_img


def _convert_input_type_range(img):
    """Convert the type and range of the input image.
    It converts the input image to np.float32 type and range of [0, 1].
    It is mainly used for pre-processing the input image in colorspace
    conversion functions such as rgb2ycbcr and ycbcr2rgb.
    Args:
        img (ndarray): The input image. It accepts:
            1. np.uint8 type with range [0, 255];
            2. np.float32 type with range [0, 1].
    Returns:
        (ndarray): The converted image with type of np.float32 and range of
            [0, 1].
    """
    img_type = img.dtype
    img = img.astype(np.float32)
    if img_type == np.float32:
        pass
    elif img_type == np.uint8:
        img /= 255.
    else:
        raise TypeError(f'The img type should be np.float32 or np.uint8, but got {img_type}')
    return img


def _convert_output_type_range(img, dst_type):
    """Convert the type and range of the image according to dst_type.
    It converts the image to desired type and range. If `dst_type` is np.uint8,
    images will be converted to np.uint8 type with range [0, 255]. If
    `dst_type` is np.float32, it converts the image to np.float32 type with
    range [0, 1].
    It is mainly used for post-processing images in colorspace conversion
    functions such as rgb2ycbcr and ycbcr2rgb.
    Args:
        img (ndarray): The image to be converted with np.float32 type and
            range [0, 255].
        dst_type (np.uint8 | np.float32): If dst_type is np.uint8, it
            converts the image to np.uint8 type with range [0, 255]. If
            dst_type is np.float32, it converts the image to np.float32 type
            with range [0, 1].
    Returns:
        (ndarray): The converted image with desired type and range.
    """
    if dst_type not in (np.uint8, np.float32):
        raise TypeError(f'The dst_type should be np.float32 or np.uint8, but got {dst_type}')
    if dst_type == np.uint8:
        img = img.round()
    else:
        img /= 255.
    return img.astype(dst_type)



import random

import numpy as np
import skimage.color as sc

import torch


def get_patch(lr, hr, ref, patch_size=108, scale=2):

    L_h, L_w = lr.shape[:2]
    L_p = patch_size

    L_x = random.randrange(L_w//4, 3*L_w//4 - L_p + 1 - 15)
    L_y = random.randrange(L_h//4, 3*L_h//4 - L_p + 1 - 15)

    H_x, H_y = scale * L_x, scale * L_y
    H_p = scale * L_p

    patch_LR = lr[L_y:L_y + L_p, L_x:L_x + L_p, :]
    patch_HR = hr[H_y:H_y + H_p, H_x:H_x + H_p, :]
    delta = random.randint(0, 30)
    patch_ref = ref[(L_y - L_h//4)*scale + delta:(L_y - L_h//4)*scale + H_p + delta, (L_x - L_w//4)*scale + delta:(L_x - L_w//4)*scale + H_p + delta, :]
    if hr.shape == lr.shape:
        return patch_LR, patch_LR, patch_ref
    return patch_LR, patch_HR, patch_ref


def set_channel(*args, n_channels=3):
    def _set_channel(img):
        if img.ndim == 2:
            img = np.expand_dims(img, axis=2)

        c = img.shape[2]
        if n_channels == 1 and c == 4:
            img = img[:, :, :3]
            img = np.expand_dims(sc.rgb2ycbcr(img)[:, :, 0], 2)
        if n_channels == 1 and c == 3:
            img = np.expand_dims(sc.rgb2ycbcr(img)[:, :, 0], 2)
        elif n_channels == 3 and c == 1:
            img = np.concatenate([img] * n_channels, 2)

        return img

    return [_set_channel(a) for a in args]


def np2Tensor(*args, rgb_range=255):
    def _np2Tensor(img):
        np_transpose = np.ascontiguousarray(img.transpose((2, 0, 1)))
        tensor = torch.from_numpy(np_transpose).float()
        tensor.mul_(rgb_range / 255)

        return tensor

    return [_np2Tensor(a) for a in args]



import collections
from functools import reduce
import ast
import argparse
from pathlib import Path
import json
import yaml
import sys
import os
import shutil

from munch import Munch
from colorama import init, Fore
init(autoreset=True)

from ._util import action_confirm

def basic_args(description=''):
    """
    fresh new training:
        python run.py train -c [config.yaml] -s [save_dir]
    resume training:
        python run.py train -s [save_dir] -r latest
    resume training with overrided config
        python run.py train -c [override.yaml] -s [save_dir] -r latest  # this will override the original config
        python run.py train -c [config.yaml] -s [save_dir] -r latest -n [new_save_dir]
    test
        python run.py test -s [save_dir] -r best
    test with overrided config
        python run.py test -s [save_dir] -r best -c config.yaml # test won't override the original config, but save the override config in test directory
        python run.py test -s [save_dir] -r best -n new_save_dir # save in a new place
    """

    args = argparse.ArgumentParser(description=description)
    args.add_argument('mode', type=str, help='running mode',
                      choices=['train', 'test', 'debug'])
    args.add_argument('-c', '--config', nargs='*', default=None, type=str,
                      help='config file(s) path (default: None)')
    args.add_argument('-r', '--resume', default=None, type=str,
                      help='resume to # checkpoint (default: None), e.g. best | latest | epoch# or a complete path')
    args.add_argument('-d', '--device', default='cuda', type=str,
                      help='indices of GPUs to enable (default: cuda)')
    args.add_argument('-s', '--save_dir', default='saved', type=str, required=True,
                      help='path of log directory (default: saved)')
    args.add_argument('-n', '--new_save_dir', default=None, type=str,
                      help='path of new log directory (default: new_saved)')
    args.add_argument('--override', default=None, type=str,
                      help='custom value to override the corrsponding key in config file.'
                      'format: "key.key=value; key2=value2", each KV pair must be seperate by a semicolon.')
    args.add_argument('--reset', action='store_true', default=False, 
                      help='remove the old log dir if exists, only used in train mode')
    args = args.parse_args()

    vars(args)['resume_dir'] = args.save_dir
    if args.new_save_dir is not None:
        args.save_dir = args.new_save_dir

    if args.mode == 'train' and args.reset:
        if os.path.exists(args.save_dir) and \
            action_confirm(Fore.RED + f'Do you really want to reset the old log?\nPath=({args.save_dir})'):
            shutil.rmtree(args.save_dir)
    
    if args.mode == 'test':
        assert args.resume is not None, 'resume cannot be None in test mode'

    if args.resume:
        if args.config and args.new_save_dir:
            cfg = read_config(args.config)
        else:
            resume_config_path = Path(args.resume_dir) / 'config.yaml'
            cfg = read_yaml(resume_config_path)
            if args.config:
                 deep_update(cfg, read_config(args.config))
    else:
        if args.config is None:
            print(Fore.RED + 'Warning: default config not founded, forgot to specify a configuration file?')
            cfg = {'engine': {}}
        else:
            cfg = read_config(args.config)

    os.makedirs(os.path.join(args.save_dir, 'history'), exist_ok=True)
    with open(os.path.join(args.save_dir, 'history', 'cmd.log'), 'a') as f:
        cmd = 'python ' + ' '.join(sys.argv) + '\n'
        f.write(cmd)

    _set_custom_args(cfg, args.override)

    return args, Munch.fromDict(cfg)


def deep_update(source, overrides):
    """
    Update a nested dictionary or similar mapping.
    Modify ``source`` in place.
    """
    for key, value in overrides.items():
        if isinstance(value, collections.Mapping) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = overrides[key]
    return source


def read_config(configs):
    cfg = {}
    # the first must be full path, can omit extension, 
    # the subseqeunt can use file name only
    base_dir = os.path.dirname(configs[0])
    for config in configs:
        # if without extension, append it
        if not config.endswith('.yaml'):
            config = config + '.yaml'
        # if file not exists, try add base_dir
        if not os.path.exists(config):
            config = os.path.join(base_dir, config)
        deep_update(cfg, read_yaml(config))
    return cfg

# ---------------------------------------------------------------------------- #
#                          Parse custom overried args                          #
# ---------------------------------------------------------------------------- #

def get_default(a:dict, b):
    return a.get(b, {})

def get_by_path(root, items):
    """Access a nested object in root by item sequence."""
    return reduce(get_default, items, root)


def set_by_path(root, items, value):
    """Set a value in a nested object in root by item sequence."""
    get_by_path(root, items[:-1])[items[-1]] = value


def _eval(val):
    try:
        val = ast.literal_eval(val)
    except:
        return val
    return val


def _remove_whitespace(array):
    return [s.strip() for s in array]


def _set_custom_args(origin: dict, override: str):
    """ --override "module.lr=0.01; engine.max_epochs=10"
    """
    if override is None:
        return
    options = _remove_whitespace(override.split(';'))
    for option in options:
        keys, value = tuple(_remove_whitespace(option.split('=')))
        keys = _remove_whitespace(keys.split('.'))
        value = _eval(value)
        set_by_path(origin, keys, value)

# ---------------------------------------------------------------------------- #
#                            Read/Write config file                            #
# ---------------------------------------------------------------------------- #


def read_json(fname):
    fname = Path(fname)
    with fname.open('rt') as handle:
        return json.load(handle)


def write_json(content, fname):
    fname = Path(fname)
    with fname.open('wt') as handle:
        json.dump(content, handle, indent=4, sort_keys=False)


def read_yaml(fname):
    fname = Path(fname)
    with fname.open('rt') as handle:
        return yaml.load(handle, Loader=yaml.FullLoader)


def write_yaml(content, fname):
    fname = Path(fname)
    with fname.open('wt') as handle:
        yaml.safe_dump(content, handle, indent=4, sort_keys=False)



import numpy as np
from skimage.metrics import peak_signal_noise_ratio, structural_similarity
from ._util import enable_batch_input, torch2numpy, bandwise, CHW2HWC

__all__ = [
    'psnr',
    'ssim',
    'sam',
    'mpsnr',
    'mssim'
]

# raw psnr, ssim, sam assume HWC format


@torch2numpy
@enable_batch_input()
@CHW2HWC
def psnr(output, target, data_range=1):
    return peak_signal_noise_ratio(target, output, data_range=data_range)


@torch2numpy
@enable_batch_input()
@CHW2HWC
def ssim(img1, img2, **kwargs):
    return structural_similarity(img1, img2, multichannel=True, **kwargs)


@torch2numpy
@enable_batch_input()
@CHW2HWC
def sam(img1, img2, eps=1e-8):
    """
    Spectral Angle Mapper which defines the spectral similarity between two spectra
    """
    tmp1 = np.sum(img1*img2, axis=2) + eps
    tmp2 = np.sqrt(np.sum(img1**2, axis=2)) + eps
    tmp3 = np.sqrt(np.sum(img2**2, axis=2)) + eps
    tmp4 = tmp1 / tmp2 / tmp3
    angle = np.arccos(tmp4.clip(-1, 1))
    return np.mean(np.real(angle))


@torch2numpy
@enable_batch_input()
@bandwise
def mpsnr(output, target, data_range=1):
    return peak_signal_noise_ratio(target, output, data_range=data_range)


@torch2numpy
@enable_batch_input()
@bandwise
def mssim(img1, img2, **kwargs):
    return structural_similarity(img1, img2, **kwargs)



# Created by Qixun Qu
# quqixun@gmail.com
# 2017/04/10
#


import numpy as np
from .affine_transform import Affine


# The number of iterations in RANSAC
ITER_NUM = 2000


class Ransac():

    def __init__(self, K=3, threshold=1):
        ''' __INIT__

            Initialize the instance.

            Input argements:

            - K : the number of corresponding points,
            default is 3
            - threshold : determing which points are inliers
            by comparing residual with it

        '''

        self.K = K
        self.threshold = threshold

    def residual_lengths(self, A, t, pts_s, pts_t):
        ''' RESIDUAL_LENGTHS

            Compute residual length (Euclidean distance) between
            estimation and real target points. Estimation are
            calculated by the given source point and affine
            transformation (A & t).

            Input arguments:

            - A, t : the estimated affine transformation calculated
            by least squares method
            - pts_s : key points from source image
            - pts_t : key points from target image

            Output:

            - residual : Euclidean distance between estimated points
            and real target points

        '''

        if not(A is None) and not(t is None):
            # Calculate estimated points:
            # pts_esti = A * pts_s + t
            pts_e = np.dot(A, pts_s) + t

            # Calculate the residual length between estimated points
            # and target points
            diff_square = np.power(pts_e - pts_t, 2)
            residual = np.sqrt(np.sum(diff_square, axis=0))
        else:
            residual = None

        return residual

    def ransac_fit(self, pts_s, pts_t):
        ''' RANSAC_FIT

            Apply the method of RANSAC to obtain the estimation of
            affine transformation and inliers as well.

            Input arguments:

            - pts_s : key points from source image
            - pts_t : key points from target image

            Output:

            - A, t : estimated affine transformation
            - inliers : indices of inliers that will be applied to refine the
            affine transformation

        '''

        # Create a Affine instance to do estimation
        af = Affine()

        # Initialize the number of inliers
        inliers_num = 0

        # Initialize the affine transformation A and t,
        # and a vector that stores indices of inliers
        A = None
        t = None
        inliers = None

        for i in range(ITER_NUM):
            # Randomly generate indices of points correspondences
            idx = np.random.randint(0, pts_s.shape[1], (self.K, 1))
            # Estimate affine transformation by these points
            A_tmp, t_tmp = af.estimate_affine(pts_s[:, idx], pts_t[:, idx])

            # Calculate the residual by applying estimated transformation
            residual = self.residual_lengths(A_tmp, t_tmp, pts_s, pts_t)

            if not(residual is None):
                # Obtain the indices of inliers
                inliers_tmp = np.where(residual < self.threshold)
                # Obtain the number of inliers
                inliers_num_tmp = len(inliers_tmp[0])

                # Set affine transformation and indices og inliers
                # in one iteration which has the most of inliers
                if inliers_num_tmp > inliers_num:
                    # Update the number of inliers
                    inliers_num = inliers_num_tmp
                    # Set returned value
                    inliers = inliers_tmp
                    A = A_tmp
                    t = t_tmp
            else:
                pass

        return A, t, inliers



from typing import Union
import numpy as np
import torch
import torch.fft as fft


def img2fft_np(img):
    dft = np.fft.fftn(img, axes=(-2, -1))
    dft_shift = np.fft.fftshift(dft)
    magnitude_spectrum = 20*np.log(np.abs(dft_shift))
    return magnitude_spectrum


def img2fft_torch(img):
    dft = fft.fftn(img, dim=(-2, -1))
    dft_shift = fft.fftshift(dft, dim=(-2, -1))
    magnitude_spectrum = 20*torch.log(torch.abs(dft_shift))
    return magnitude_spectrum


def img2fft(img: Union[torch.Tensor, np.ndarray]):
    """ img's shape [...,W,H] """

    if isinstance(img, torch.Tensor):
        return img2fft_torch(img)
    elif isinstance(img, np.ndarray):
        return img2fft_np(img)
    else:
        raise ValueError('unsupported data type ' + str(type(img)))



import functools
import numpy as np

FORMAT_HWC = 'HWC'
FORMAT_CHW = 'CHW'
DATA_FORMAT = FORMAT_HWC


def set_data_format(format):
    if format != FORMAT_HWC and format != FORMAT_CHW:
        raise ValueError('Invalid data format, choose from '
                         'torchlight.metrics.HWC or torchlight.metrics.CHW')
    global DATA_FORMAT
    DATA_FORMAT = format


def CHW2HWC(func):
    @functools.wraps(func)
    def warpped(output, target, *args, **kwargs):
        if DATA_FORMAT == FORMAT_CHW:
            output = output.transpose(1, 2, 0)
            target = target.transpose(1, 2, 0)
        return func(output, target, *args, **kwargs)
    return warpped


def torch2numpy(func):
    @functools.wraps(func)
    def warpped(output, target, *args, **kwargs):
        if not isinstance(output, np.ndarray):
            output = output.detach().cpu().numpy()
            target = target.detach().cpu().numpy()
        return func(output, target, *args, **kwargs)
    return warpped


def bandwise(func):
    @functools.wraps(func)
    def warpped(output, target, *args, **kwargs):
        if DATA_FORMAT == FORMAT_CHW:
            C = output.shape[-3]
            total = 0
            for ch in range(C):
                x = output[ch, :, :]
                y = target[ch, :, :]
                total += func(x, y, *args, **kwargs)
            return total / C
        else:
            C = output.shape[-1]
            total = 0
            for ch in range(C):
                x = output[:, :, ch]
                y = target[:, :, ch]
                total += func(x, y, *args, **kwargs)
            return total / C
    return warpped


def enable_batch_input(reduce=True):
    def inner(func):
        @functools.wraps(func)
        def warpped(output, target, *args, **kwargs):
            if len(output.shape) == 4:
                b = output.shape[0]
                out = [func(output[i], target[i]) for i in range(b)]
                if reduce:
                    return sum(out) / len(out)
                return out
            return func(output, target, *args, **kwargs)
        return warpped
    return inner



import numpy as np
import random

from ._util import LockedIterator


class Compose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, data):
        out = data
        for transform in self.transforms:
            out = transform(out)
        return out


class SequentialSelect(object):
    def __pos(self, n):
        i = 0
        while True:
            # print(i)
            yield i
            i = (i + 1) % n

    def __init__(self, transforms):
        self.transforms = transforms
        self.pos = LockedIterator(self.__pos(len(transforms)))

    def __call__(self, img):
        out = self.transforms[next(self.pos)](img)
        return out


class MinMaxNormalize:
    def __call__(self, array):
        amin = np.min(array)
        amax = np.max(array)
        return (array - amin) / (amax - amin)


class CenterCrop:
    def __init__(self, size):
        self.cropx = size[0]
        self.cropy = size[1]

    def __call__(self, img):
        x, y = img.shape[-2], img.shape[-1]
        startx = x//2-(self.cropx//2)
        starty = y//2-(self.cropy//2)
        return img[..., startx:startx+self.cropx, starty:starty+self.cropy]


class RandCrop:
    def __init__(self, size):
        self.cropx = size[0]
        self.cropy = size[1]

    def __call__(self, img):
        x, y = img.shape[-2], img.shape[-1]
        x1 = random.randint(0, x - self.cropx)
        y1 = random.randint(0, y - self.cropy)
        return img[..., x1:x1+self.cropx, y1:y1+self.cropy]





from setuptools import setup, find_packages

setup(
    name='torchlights',
    packages=find_packages(),
    version='0.3.1.1',
    install_requires=['munch', 'colorama', 'readchar', 'tqdm', 'qqdm', 'pyyaml', 'colorlog'],
    include_package_data=True
)


import random

class RandCrop:
    def __init__(self, img_size, crop_size):
        self.cropx = crop_size[0]
        self.cropy = crop_size[1]
        x, y = img_size[0], img_size[1]
        self.x1 = random.randint(0, x - self.cropx)
        self.y1 = random.randint(0, y - self.cropy)
        
    def __call__(self, img):
        return img[..., self.x1:self.x1+self.cropx, self.y1:self.y1+self.cropy]




import numpy as np
from skimage.util import random_noise

from ._util import LockedIterator


class GaussianNoise(object):
    def __init__(self, sigma):
        self.sigma = sigma

    def __call__(self, img):
        img_L = img + np.random.normal(0, self.sigma, img.shape)
        return img_L


class GaussianNoiseBlind(object):
    """add blind gaussian noise to the given numpy array (B,H,W)"""

    def __pos(self, n):
        i = 0
        while True:
            yield i
            i = (i + 1) % n

    def __init__(self, sigmas):
        self.sigmas = np.array(sigmas) / 255.
        self.pos = LockedIterator(self.__pos(len(sigmas)))

    def __call__(self, img):
        noise = np.random.randn(*img.shape) * self.sigmas[next(self.pos)]
        return img + noise


class GaussianNoiseBlindv2(object):
    """add blind gaussian noise to the given numpy array (B,H,W)"""

    def __init__(self, min_sigma, max_sigma):
        self.min_sigma = min_sigma
        self.max_sigma = max_sigma

    def __call__(self, img):
        noise = np.random.randn(
            *img.shape) * np.random.uniform(self.min_sigma, self.max_sigma) / 255
        return img + noise


class GaussianNoiseNoniid(object):
    """add non-iid gaussian noise to the given numpy array (B,H,W)"""

    def __init__(self, sigmas):
        self.sigmas = np.array(sigmas) / 255.

    def __call__(self, img):
        bwsigmas = np.reshape(self.sigmas[np.random.randint(
            0, len(self.sigmas), img.shape[0])], (-1, 1, 1))
        noise = np.random.randn(*img.shape) * bwsigmas
        return img + noise


class ImpulseNoise(object):
    """add impulse noise to the given numpy array (B,H,W)"""

    def __init__(self, amounts, s_vs_p=0.5):
        self.amounts = np.array(amounts)
        self.s_vs_p = s_vs_p

    def __call__(self, img, bands):
        # bands = np.random.permutation(range(img.shape[0]))[:self.num_band]
        bwamounts = self.amounts[np.random.randint(
            0, len(self.amounts), len(bands))]
        for i, amount in zip(bands, bwamounts):
            img[i, ...] = random_noise(
                img[i, ...], mode='s&p', amount=amount, salt_vs_pepper=self.s_vs_p)
        return img


class StripeNoise(object):
    """add stripe noise to the given numpy array (B,H,W)"""

    def __init__(self, min_amount, max_amount):
        assert max_amount > min_amount
        self.min_amount = min_amount
        self.max_amount = max_amount

    def __call__(self, img, bands):
        B, H, W = img.shape
        # bands = np.random.permutation(range(img.shape[0]))[:len(bands)]
        num_stripe = np.random.randint(
            np.floor(self.min_amount*W), np.floor(self.max_amount*W), len(bands))
        for i, n in zip(bands, num_stripe):
            loc = np.random.permutation(range(W))
            loc = loc[:n]
            stripe = np.random.uniform(0, 1, size=(len(loc),))*0.5-0.25
            img[i, :, loc] -= np.reshape(stripe, (-1, 1))
        return img


class DeadlineNoise(object):
    """add deadline noise to the given numpy array (B,H,W)"""

    def __init__(self, min_amount, max_amount):
        assert max_amount > min_amount
        self.min_amount = min_amount
        self.max_amount = max_amount

    def __call__(self, img, bands):
        B, H, W = img.shape
        # bands = np.random.permutation(range(img.shape[0]))[:len(bands)]
        num_deadline = np.random.randint(
            np.ceil(self.min_amount*W), np.ceil(self.max_amount*W), len(bands))
        for i, n in zip(bands, num_deadline):
            loc = np.random.permutation(range(W))
            loc = loc[:n]
            img[i, :, loc] = 0
        return img


class MixedNoise(object):
    """add mixed noise to the given numpy array (B,H,W)
    Args:
        noise_bank: list of noise maker (e.g. AddNoiseImpulse)
        num_bands: list of number of band which is corrupted by each item in noise_bank"""

    def __init__(self, noise_bank, num_bands):
        assert len(noise_bank) == len(num_bands)
        self.noise_bank = noise_bank
        self.num_bands = num_bands

    def __call__(self, img):
        B, H, W = img.shape
        all_bands = np.random.permutation(range(B))
        pos = 0
        for noise_maker, num_band in zip(self.noise_bank, self.num_bands):
            if 0 < num_band <= 1:
                num_band = int(np.floor(num_band * B))
            bands = all_bands[pos:pos+num_band]
            pos += num_band
            img = noise_maker(img, bands)
        return img



import cv2
import imageio
import numpy as np
import matplotlib.pyplot as plt


def imread(path):
    """ a wrapper of imageio.imread, which removes the meta data.
    """
    img = imageio.imread(path)
    img = np.array(img)
    return img


def imwrite(img, path):
    """ a wrapper of imageio.imwrite 
    """
    imageio.imwrite(img, path)


def imshow(img):
    plt.imshow(img)


def hwc2chw(img):
    return img.transpose(2, 0, 1)


def uint2float(img):
    return img.astype('float') / 255


def float2uint(img):
    return (img * 255).astype('uint8')


def rgb2bgr(img):
    """ input [H,W,C] """
    return img[:, :, ::-1]


def usm_sharp(img, weight=0.5, radius=50, threshold=10):
    """USM sharpening.
    Input image: I; Blurry image: B.
    1. sharp = I + weight * (I - B)
    2. Mask = 1 if abs(I - B) > threshold, else: 0
    3. Blur mask:
    4. Out = Mask * sharp + (1 - Mask) * I
    Args:
        img (Numpy array): Input image, HWC, BGR; float32, [0, 1].
        weight (float): Sharp weight. Default: 1.
        radius (float): Kernel size of Gaussian blur. Default: 50.
        threshold (int):
    """
    if radius % 2 == 0:
        radius += 1
    blur = cv2.GaussianBlur(img, (radius, radius), 0)
    residual = img - blur
    mask = np.abs(residual) * 255 > threshold
    mask = mask.astype('float32')
    soft_mask = cv2.GaussianBlur(mask, (radius, radius), 0)

    sharp = img + weight * residual
    sharp = np.clip(sharp, 0, 1)
    return soft_mask * sharp + (1 - soft_mask) * img



import torch

def accuracy(output, target):
    with torch.no_grad():
        pred = torch.argmax(output, dim=1)
        assert pred.shape[0] == len(target)
        correct = 0
        correct += torch.sum(pred == target).item()
    return correct / len(target)


def top_k_acc(output, target, k=3):
    with torch.no_grad():
        pred = torch.topk(output, k, dim=1)[1]
        assert pred.shape[0] == len(target)
        correct = 0
        for i in range(k):
            correct += torch.sum(pred[:, i] == target).item()
    return correct / len(target)



from .degrade import GaussianBlur, UniformBlur, KFoldDownsample, UniformDownsample, GaussianDownsample, HSI2RGB, Resize, Upsample
from .noise import GaussianNoise, GaussianNoiseBlind, GaussianNoiseBlindv2, ImpulseNoise, StripeNoise, DeadlineNoise, MixedNoise
from .general import Compose, MinMaxNormalize, CenterCrop, RandCrop